# This module contains classes for a minesweeper game's GUI
module MineSweeperConnectionModule

require 'rubygems'
require 'thread'
require 'monitor'
require 'MineSweeperUtilModule'
require 'socket'
require 'SocketPackets'
include Socket::Constants

# This class implements a client that can handle a {MineSweeperConnectionModule::MineSweeperServer server} server
# @defaultreturn [NotImportant] Not important value
class MineSweeperClient
 # @return [MineSweeperModule::MineSweeperApplication] The owner of the object
 attr_reader :owner
 
 # Sets the owner of the object 
 # 
 # @param [MineSweeperModule::MineSweeperApplication] owner the owner
 # @return [Boolean] true if success
 def initialize(owner)
  @owner = owner
  true
 end
 
 # Start the client. If the server is not avaliable the client tells it to the owner 
 def start
   @mutex = Mutex.new
   @buffer = ""
   @sock = Socket.new(MineSweeperConnectionModule::AF_INET,MineSweeperConnectionModule::SOCK_STREAM,0)
   sockaddr = "127.0.0.1:42424"
   @host,@port = sockaddr.split(":")
   sockaddr = Socket.sockaddr_in(@port,@host)
   hiba = true
   while (hiba)
    sleep 0.1
    begin
    @sock.connect_nonblock(sockaddr)
    rescue Errno::EINPROGRESS
	 hiba = true
	 next
	rescue Errno::EALREADY
	 hiba = true
	 next
	rescue Errno::EISCONN
	 hiba = false
    rescue Errno::ECONNREFUSED
     @owner.notConnected
    end
    hiba = false
   end
   begin
    @sock.send_packet("connected")
    @eof = false
    @thread = Thread.new {
     p "started"
     loop do
       @mutex.lock
       ret = IO::select([@sock],nil,nil,0.3)
       unless ret.nil?
         if @sock.eof?
           @sock.close
           @sock = nil
           @mutex.unlock
		   @owner.notConnected
           @thread.exit
           break
         end
         msg = @sock.recv_packet()
         @buffer += msg
         loop do
           at = @buffer.index("\n")
           if at.nil?
             break
           end
           data = @buffer[0..at-1]
           @buffer = @buffer[at+1..-1]
           parse_message(data)
         end
       end
       @mutex.unlock
     end
    }
   rescue Errno::ENOTCONN
    @owner.notConnected
   rescue Errno::EPIPE
    @owner.notConnected
   end
 end
 
  # Handle the received command
  # 
  # @param [String] msg the message that we will parse
  def parse_message(msg)
   if (msg.match('^shutdown#*')) then @sock.close; @sock = nil; end
   if (msg.match('^msg#'))
    data = msg.split('#')
	data.shift
	opp = data.shift
	sender = data.shift
	tmp = data.join('#')
	@Table[opp].chat = sender + "->" + tmp
   end
   if (msg.match('^request#game#*')) then @owner.doGameRequest(msg.split('#')[2]) unless @owner.nil?; end
   if (msg.match('^request#failed#*')) then @owner.doFailedRequest(msg.split('#')[2]) unless @owner.nil?; end
   if (msg.match('^game#reject#*')) then @owner.doGameReject(msg.split('#')[2]) unless @owner.nil?; end
   if (msg.match('^game#closed#*')) then @owner.doGameClosed(msg.split('#')[2]) unless @owner.nil?; end
   if (msg.match('^table#*')) then data = msg.split('#'); @owner.startGame(data[1],data[2],data[3]) unless @owner == nil end
   if (msg.match('^joined#*')) then @owner.joined(msg.split('#')[1]) end
   if (msg.match('^left#*')) then @owner.left(msg.split('#')[1]) end
   if (msg.match('^globmsg#*')) then data = msg.split('#'); @owner.gotGlobMsg(data[1],data[2..data.size-1]) end
   if (msg.match('^list#*')) then data = msg.split('#'); data.shift; @owner.setUserList(data) unless @owner == nil end
   if (msg.match('^login#ok#*')) then @owner.doLogin(msg.split('#')[2]) unless @owner == nil end
   if (msg.match('^login#denied#exists*')) then @owner.onNickExists unless @owner.nil? end
   if (msg.match('^hit#*')) then  data = msg.split('#'); @owner.doHit(data[1],data[2],data[3]) end
   if (msg.match('^win#*')) then @owner.win(msg.split('#')[1]) end
  end
  
 # Send message to the server
 # 
 # @param [String] msg The message that we will send
 def send(msg)
  @mutex.lock
   if (!@sock.nil?) then @sock.send_packet(msg) end
  @mutex.unlock
 end
 
 # Stops the client
 def stop
  @mutex.lock
   @sock.close unless @sock.nil?
   @sock = nil
  @mutex.unlock
 end

end

# This class implements a server that handle {MineSweeperConnectionModule::MineSweeperClient clients}
# @defaultreturn [NotImportant] Not important value
class MineSweeperServer
  attr_reader :started
  def initialize
   @started = false
  end
  
  # ClientInfo holds information about each client connection to the server
  # making it easier to track information, and use that throughout the
  # server process.
  #
  # @attr [Array]  opps The client's opponents
  # @attr [Array]  sended Sended game requests(nicklist)
  # @attr [String] name Nickname
  # @attr [String] host Client's hostname
  # @attr [Integer] port Client's port number
  # @attr [Socket] sock Client's socket
  # @attr [String] buffer Client's buffer for messages
  ClientInfo = Struct.new( :opps, :sended, :name, :host, :port, :sock, :buffer)
  
						   
						   
  #This method initialize the server. You must call it before {MineSweeperServer#run run method}
  def init
    @port = 42424       # Port we're listening on
    @mutex = Mutex.new # For synchronization, so there's no corruption of memory
    @clients = []      # For the sockets that are connected
  end
  
  # This function sets up the socket for listening to incoming connections, as
  # well as starting the Thread that will do the underline work for polling for
  # incoming connections as well as incoming data.
  def run
    @socket = Socket.new(MineSweeperConnectionModule::AF_INET,MineSweeperConnectionModule::SOCK_STREAM,0)
    @socket.setsockopt(MineSweeperConnectionModule::SOL_SOCKET,MineSweeperConnectionModule::SO_REUSEADDR, 1)
    @socket.bind(Socket.sockaddr_in(@port,"0.0.0.0"))
    @socket.listen(100)
    start_server_thread()
    @started = true
  end
  
  
  # This is the actual clean up function that will shutdown the Server, and
  # disconnect any connected clients.  When this is executed, there should
  # already have a Mutex lock done, so that we can safely clean everything up.
  def stop
    Thread.kill(@thread) unless @thread.nil?
    @timer.stop
    
    @clients.each do |ci|
	  ci.sock.send_packet("globmsg#Server#Closed")
      ci.sock.send_packet("shutdown#")
      ci.sock.close
    end
    @clients = []
    @socket.close unless @socket.nil?
    @socket = nil
	@count = 0
  end
  
  
  # Return dump message
  # @return [Array(String,Integer)] The dump message in [status,client count] format
  def dump; @mutex.lock; res = [ (@started and !@socket.nil? and @thread.alive?) ? "started" : "stopped" ,@count]; @mutex.unlock; res end
  
  # Return server running status
  # @return [Boolean] Indicates that the server is running
  def isStarted; @started;  end
  
  # Send a global chat message for all clients
  # @param [String] msg The message
  def sendAll(msg)
   @mutex.lock
    @clients.each do |ci|
      ci.sock.send_packet("globmsg#Server##{msg}")
    end
   @mutex.unlock
  end
  
  # Return a client status
  # @param [String] nick The client's nickname
  # @return [String] The player state
  def state(nick)
   index = @clients.index {|x| x.name.eql?(nick) }
   index.nil? ? "notloggedin" : ( (@clients[index].opps.size > 0 ) ? "ingame#" + @clients[index].opps.join('#') :"avaliable")
  end
  
  
  private
  
  # This method, will run through the collected client connections that show that
  # they have data available in their buffer.  It will first check to see if the
  # socket has been closed by utilizing the IO's eof? method.  If there is an end
  # of file detected, it means the client connection has been closed.  It will then
  # remove the client connection from our pool, and send the rest of the clients
  # that the user has indeed disconnected from the server.
  # After confirming that the connection hasn't been closed, it will check to
  # see what data is available, append it to the client's buffer, where it will
  # then process the data, for the LF (0x0A) termination, saying that this is the
  # end of the message, and then process that data through {MineSweeperServer#parse_message parse_message}
  # @param [Array(ClientInfo)] aci The array with {MineSweeperServer::ClientInfo ClientInfo} structures
  def check_clients(aci)
    aci.each do |ci|
      if (ci.sock.eof?)
	    clearClient(ci)
        ci.sock.close
        @clients.delete(ci)
        @count -= 1
      else
        msg = ci.sock.recv_packet()
        ci.buffer += msg
        loop do
          at = ci.buffer.index("\n")
          if at.nil?
            break
          end
          data = ci.buffer[0..at-1]
          ci.buffer = ci.buffer[at+1..-1]
          parse_message(ci,data)
        end
      end
    end
  end
  
  # This is the meat and bones of the server, where it will continiously poll for
  # data in a seperate Thread, that will utilize the above methods for processing
  # the incoming connections, and data being inputed from clients.
  def start_server_thread()
    @thread = nil
    while (@thread.nil?)
     @thread = Thread.new do
       @timer = Timer.new(0.25) {Thread.pass}
       @count = 0
       loop do # We just run a continious loop.  on_stop() will kill this thread, and stop the timer.
         begin	# We run here, just incase there's some kind of error
           @mutex.lock # We lock the data, so there's no corruption between threads
           socks = poll() # We run a poll for any incoming data, or any new incoming connections
           unless socks.nil?
             ci = check_server(socks[0]) # Check to see if we have any incoming connections
             check_clients(ci)	# Check for client activity
           end
           @mutex.unlock # We can now release it, as we've done everything needed here
         rescue => e  # We have an error, so we need to process it.
           msg = ["An error has occured!"]
           msg << "#{e.backtrace.delete_at(0)}: #{e} (#{e.class})"
           e.backtrace.each do |line|
             msg << line
           end
           msg.each do |line|
             p line
           end
         end
       end
     end
    end
  end

  # This function collects the sockets for clients, as well as the server
  # socket to poll for incoming connections or data from these sockets.
  # This returns three arrays of sockets from [read,write,error/oob]
  def poll
    pollers = [@socket]
    @clients.each do |ci|
      pollers << ci.sock
    end
    IO::select(pollers,nil,nil,0.3)
  end
  
  # This is the core of the processing for the server side.  It will check to
  # see if our server socket is in the return for {MineSweeperServer#poll poll}, if it's
  # there, then there's an incoming connection, and will run the accepting of
  # the new client connection. It will then remove the server socket from the array
  # so that it's not processed, as well as converting all the remaining sockets
  # if any, back into {MineSweeperServer::ClientInfo ClientInfo} structures, there by returning them to be
  # processed by the {MineSweeperServer#check_clients check_clients method}
  # @param [Array(Socket)] socks The array with sockets that we will handle
  # @return [Array(ClientInfo)] new_ci The new client array
  def check_server(socks)
    if socks.index(@socket)
      socks.delete(@socket) # Don't need it in socks
      sock, sockaddr = @socket.accept
      sockaddr = Socket.unpack_sockaddr_in(sockaddr)
      ci = ClientInfo.new([],[],"",sockaddr[1],sockaddr[0],sock,"")
      @count += 1
      @clients << ci
      sock.puts("hello")
    end
    new_ci = []
    socks.each do |sock|
      @clients.each do |ci|
        if (!sock.closed? and ci.sock == sock)
          new_ci << ci
        end
      end
    end
    new_ci
  end

  
  # Parse the current message
  # @param [ClientInfo] ci The client's ClientInfo that sended the message
  # @param [String] msg The received message
  def parse_message(ci,msg)
   if (msg.match('^login#')) then data = msg.split('#');data.shift;loginRequire(ci,data.join('#')) end
   if (msg.match('^logout#')) then logoutRequire(ci) end
   if (msg.match('^exit#$')) then exitRequire(ci);end
   if (msg.match('^request#game#*')) then startRequire(ci,msg.split('#')[2]) end
   if (msg.match('^globmsg#')) then sendPublicMessage(ci,msg.split('#')) end
   if (msg.match('^msg#')) then sendPrivateMessage(ci,msg.split('#')) end
   if (msg.match('^game#accept#*')) then startGame(ci,msg.split('#')[2]) end
   if (msg.match('^game#reject#*')) then rejectGame(ci,msg.split('#')[2]) end
   if (msg.match('^game#closed#*')) then closeGame(ci,msg.split('#')[2]) end
   if (msg.match('^game#lost#*')) then data = msg.split('#'); endGame(ci,data[2]) end
   if (msg.match('^hit#*')) then doHit(ci,msg.split('#')[1..3]) end
  end
  
  # Send the active clients's nick names to a client
  # @param [ClientInfo] s The destination ClientInfo
  def sendList(s)
   list = []
   @clients.each do |c|
    if (!c.name.eql?("") and !c.name.nil? and !c.name.eql?(s.name)) then list << c.name end
   end
   msg = "list##{list.join('#')}"
   s.sock.send_packet(msg)
  end
  
  # Send a hit message if it is valid
  # @param [ClientInfo] s The sender
  # @param [Array(String)] data The received data in [opponent,x coordinate,y coordinate] format
  def doHit(s,data)
   index = @clients.index { |x| x.name.eql?(data[0])}
   if (!index.nil?)
    to = @clients[index]
	to.sock.send_packet("hit##{s.name}##{data[1]}##{data[2]}")
	s.sock.send_packet("hit##{to.name}##{data[1]}##{data[2]}")
   end
  end
  
  # Send to the client opponent that it won the game
  # @param [ClientInfo] s The sender
  # @param [String] opp Opponent's nickname
  def endGame(s,opp)
   index = s.opps.index { |x| x.eql?(opp) }
   if (!index.nil?)
    index = @clients.index { |x| x.name.eql?(opp) }
	if (!index.nil?)
     to = @clients[index]
	 to.sock.send_packet("win##{s.name}")
	 s.opps.delete(to.name)
	 to.opps.delete(s.name)
	end
   end
  end
  
  # Send to the opponents that the game was closed
  # @param [ClientInfo] s The sender
  # @param [String] opp Opponent's nickname  
  def closeGame(s,opp)
   index = s.opps.index { |x| x.eql?(opp) }
   if (!index.nil?)
    index = @clients.index { |x| x.name.eql?(opp) }
	if (!index.nil?)
     to = @clients[index]
	 to.sock.send_packet("game#closed##{s.name}")
	 s.opps.delete(to.name)
	 to.opps.delete(s.name)
	end
   end
  end

  # Send a reject message to the opponent
  # @param [ClientInfo] s The sender
  # @param [String] opp Opponent's nickname  
  def rejectGame(s,opp)
   index = @clients.index { |x| x.name.eql?(opp) }
   if (!index.nil?)
    to = @clients[index]
	to.sock.send_packet("game#reject##{s.name}")
	to.sended.delete(s.name)
   end
  end
  
  # Send a new gametable to clients
  # @param [ClientInfo] s Player one
  # @param [String] opp Player two 
  def startGame(s,opp)
   index = @clients.index { |x| x.name.eql?(opp) }
   if (!index.nil?)
    table = MineSweeperUtilModule::Utils.TableToString(MineSweeperUtilModule::Utils.generate(16,31),false);
    to = @clients[index]
	to.opps << s.name
	s.opps << to.name
	to.sock.send_packet("table##{s.name}#1##{table}")
	s.sock.send_packet("table##{opp}#0##{table}")
	to.sended.delete(s.name)
   end
  end

 # Send private message
 # @param [ClientInfo] s Sender
 # @param [Array(String)] data The opponent's nick and the message in [nick,message format]
 def sendPrivateMessage(s,data)
  if (!s.name.nil?)
   data.shift
   to = @clients[@clients.index {|x| x.name.eql?(data[0])} ]
   data.shift
   msg = data.join('#')
   to.sock.send_packet("msg##{s.name}##{s.name}##{msg}")
   s.sock.send_packet("msg##{to.name}##{s.name}##{msg}")
  end
 end
 
 # Send public message
 # @param [ClientInfo] sender Sender
 # @param [String] msg The message
 def sendPublicMessage(sender,msg)
  if (!sender.name.nil?)
   msg.shift
   @clients.each do |ci|
    ci.sock.send_packet("globmsg##{sender.name}##{msg.join('#')}")
   end
  end

 end
 
 # Handle client's exit. Send a message to others that client was left
 # and clear game assignments and sended requests
 # @param [ClientInfo] ci The exiting client
 def clearClient(ci)
  if (!ci.name.nil?)
   @clients.each do |c|
    c.sock.send_packet("left##{ci.name}") unless c.eql?(ci)
   end
   ci.opps.each do |opp|
    index = @clients.index { |x| x.name.eql?(opp) }
	if (!index.nil?)
	 @clients[index].opps.delete(ci.name)
	 @clients[index].sended.delete(ci.name)
	end
   end
  end
 end
 
 # Handle client exit from server
 # @param [ClientInfo] ci The exited client
 def exitRequire(ci)
  ci.sock.close
  @clients.delete(ci)
  @count -= 1
  clearClient(ci)
 end
 
 # Handle client logout from server
 # @param [ClientInfo] ci The client that logged out
 def logoutRequire(s)
  if (!s.name.nil?)
   clearClient(s)
   s.name = nil
  end
 end
 
 # Handle client login to server. This validates the specified nick
 # @param [ClientInfo] s The client that want to log in
 # @param [String] name The nickname to validate
 def loginRequire(s,name)
  good   = name.match('^[a-zA-Z0-9]+$');
  exists = @clients.index { |x| x.name.eql?(name) }
  if (exists.nil?)
   if (good)
	s.sock.send_packet("login#ok##{name}")
	@clients.each do |ci|
      if (!(ci == s) ) then ci.sock.send_packet("joined##{name}") end
    end
	s.name = name;
	sendList(s)
   else
    s.sock.send_packet('login#denied#bad')
   end
  else
   s.sock.send_packet('login#denied#exists')
  end
 end

 # Handle game start require. Find the opponent and tell to the client
 # if it has not found or send the request to it if it has found.
 # @param [ClientInfo] s The client that want to play
 # @param [String] opp The nickname of opponent
 def startRequire(s,opp)
  index = @clients.index { |x| x.name.eql?(opp)}
  if (index.nil?)
   s.sock.send_packet("request#failed#notexists")
  else
   to = @clients[index]
   if (to.sock.closed?)
    s.sock.send_packet("request#failed#notexists")
   else
    if (!s.sended.include?(to.name) and !s.opps.include?(to.name))
	 s.sended << to.name
     to.sock.send_packet("request#game##{s.name}")
	end
   end
  end
 end
end

# This is a simple timer class.
# @defaultreturn [NotImportant] Not important value
class Timer

 # Start the timer with the interval and when the times up 
 # this will call the handler and start a new cycle
 def initialize(interval,*args, &handler)
  raise ArgumentError, "Illegal interval" if interval < 0
  extend MonitorMixin
  @run = true
  @th = Thread.new do
   t = Time.now
   while run?
    t += interval
    (sleep(t - Time.now) rescue nil) and
    handler.call(*args) rescue nil
	Thread.pass
   end
  end
 end

 # Stops the timer
 def stop
  synchronize do
   @run = false
  end
  @th.join
 end

 private
 
 # Indicates the timer status
 def run?
  synchronize do
   @run
  end
 end
end
end
