# This module contains the classes for mine sweeper game logic
module MineSweeperUtilModule

# This class implements the mine sweeper game
# @defaultreturn [NotImportant] Not important value
class MineSweeperGame
 # Represent a field of table
 # @attr [Integer] type Type of field(0..9)
 # @attr [Boolean] visible True if the field is visible
 Field = Struct.new(:type,:visible)
 
 # @return [Integer] The size of game table
 attr_reader :size
 # @return [Integer] The number of hidden mines
 attr_reader :hiddenMines
 # @return [Integer] The number of initial mines
 attr_reader :mines
 # @return [Integer] True if the current player is the active
 attr_reader :active
 # @return [String] The name of the game
 attr_reader :name
 # @return [Array(Array(Field))] The game table
 attr_reader :data
 # @return [Integer] Score of player
 attr_reader :score
 # @return [Integer] Score of opponent
 attr_reader :oppscore
 # @return [String] Game status(nothing,ingame,win,loose)
 attr_reader :state
 
 # Initialize the game
 # @param [String] owner The name of the game
 # @return [MineSweeperGame] The new game object
 def initialize(name = nil)
  @name = name
  @data = []
  @size = 0
  @active = false
  @state = "nothing"
 end
 
 # Perform a hit to a field
 # @param [Integer] i The x coordinate
 # @param [Integer] j The y coordinate
 def hit i,j
  if (!@data[i][j].visible) then explode(i,j,0) end
 end 
 
 # Pretty prints the table to console
 def printTable
  @data.each do |line|
   line.each do |field|
    print field[:visible]? field[:type] : "X", " "
   end
   print "\n"
  end
 end
 
 # Start a game
 # @param [Boolean] True if we will start the game
 def start(red)
  @active = red
  @state = "ingame"
  @score = 0
  @oppscore = 0
 end
 
 # @return [Boolean] True if the game state is ingame
 def isStarted
  @state.eql?("ingame")
 end

 # Sets the game table
 # @param [String] tablestring String representation of table
 # @param [Integer] n The size of the table
 # @return [Boolean] True if success
 def setTable(tablestring,n)
  @hiddenMines = 0
  @size = n
  @data = Array.new(n){Array.new(n){Field.new(0,false)}}
  (0..n-1).each do |i|
   (0..n-1).each do |j|
    @data[i][j][:type] = (tablestring[i*n+j]-48).to_i
	if @data[i][j][:type] == 9 then @hiddenMines += 1 end
   end
  end
  @mines = @hiddenMines
  true
 end
 
 # Gets the game table's string representation without hiding
 # @return [String] The table's string representation
 def getTable
  MineSweeperUtilModule::Utils.TableToString(@data,false)
 end

 # Gets the game table's string representation with hiding
 # @return [String] The table's string representation
 def getTableWithMask
  MineSweeperUtilModule::Utils.TableToString(@data,true)
 end
 # Gets a field from table
 # @param [Integer] i The x coordinate
 # @param [Integer] j The y coordinate
 # @return [Field] nil if the indexes are wrong and a field if not
 def get(i,j)
  if (!@data.nil? and i<@size and i>=0 and j<@size and j>=0) 
   return @data[i][j]
  else 
   return nil
  end
 end 
 
 private
 
 # Do the exploding and mine count
 # @param [Integer] i The x coordinate
 # @param [Integer] j The y coordinate
 # @param [Integer] n Indicate that is the first call
 def explode(i,j,n)
  if (i >= 0 and j >= 0 and i < @size and j < @size)
   if (!@data[i][j].visible)
    @data[i][j].visible = true
   end
  
   (i-1..i+1).each do |k|
    (j-1..j+1).each do |l|
     if (k >= 0 and l >= 0 and k < @size and l < @size and !@data[k][l][:visible] and @data[i][j][:type] == 0) then explode(k, l, 1); end
    end
   end
   
   if (@data[i][j].type == 9)
    if (@active)
	 @data[i][j].type = 10
	 @score = @score + 1
	else
	 @data[i][j].type = 11
	 @oppscore = @oppscore + 1
	end
	if ( @score > (@mines / 2)) then @state = "win"; @active = false end
	if ( @oppscore > (@mines / 2)) then @state = "loose"; @active = false end
	@hiddenMines -= 1
   else 
    if (n == 0) then @active = !@active end
   end
  end
 end 
end

# This class contains static methods to generate tables and transform a table
# to string
class Utils
 # Represent a field of table
 # @attr [Integer] type Type of field(0..9)
 # @attr [Boolean] visible True if the field is visible
 Field = Struct.new(:type,:visible)
 
 # This method generate a valid game table
 # @param [Integer] n The table size
 # @param [Integer] count The count of mines
 # @return [Array(Array(Field))] The generated table
 def Utils.generate(n, count) 
   cnt = 0
   table = Array.new(n){Array.new(n){Field.new(0,false)}}
   while cnt < count
    newl = rand(n); newc = rand(n)
    if table[newl][newc][:type] != 9
     cnt = cnt.next 
	 table[newl][newc][:type] = 9
     (newl - 1..newl + 1).each do |i|
      (newc - 1 .. newc + 1).each do |j|
       if ((i >= 0 && j >= 0 && i< n && j < n) && table[i][j][:type] != 9) then table[i][j][:type] += 1 end
	  end
	 end
    end
   end
   table
  end
  
 # Transform a game table to string
 # @param [Array(Array(Field))] table The table
 # @param [Boolean] hide True if we want to hide hidden fields
 # @return [String] The table's string representation
 def Utils.TableToString(table,hide) 
  msg = ""
  table.each do |r|
   r.each do |i|
    if hide then
	 if i[:visible] then msg += i[:type].to_s else msg += 'X' end
	else
	 msg += i[:type].to_s
	end
   end
  end
  msg
 end
end

end